#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include <string.h>
#include <libintl.h>
#include "config.h"

static double
constant(char *name, int arg)
{
	errno = 0;
	if (strEQ(name, "LC_CTYPE")) return LC_CTYPE;
	if (strEQ(name, "LC_NUMERIC")) return LC_NUMERIC;
	if (strEQ(name, "LC_COLLATE")) return LC_COLLATE;
	if (strEQ(name, "LC_MONETARY")) return LC_MONETARY;
	if (strEQ(name, "LC_MESSAGES")) return LC_MESSAGES;
	if (strEQ(name, "LC_ALL")) return LC_ALL;
	errno = EINVAL;
	return 0;
}

char *current_domain = NULL;
HV *domains_in_utf8 = NULL;

int
is_domain_utf8(const char *domain)
{
	return domain && domains_in_utf8 && hv_exists(domains_in_utf8, domain, strlen(domain));
}

#define ANY_MISSING 0

#ifndef HAVE_DGETTEXT
/* if dgettext is not there, neither will
   dcgettext, dngettext and dcngettext be.
   But only deal with dngettext and
   dcngettext if they will not be dealt
   with later because of ngettext */
#define dgettext(a,b) not_here("dgettext")
#define dcgettext(a,b,c) not_here("dcgettext")
#ifdef HAVE_NGETTEXT
#define dngettext(a,b,c,d) not_here("dngettext")
#define dcngettext(a,b,c,d,e) not_here("dcngettext")
#endif
#undef ANY_MISSING
#define ANY_MISSING 1
#endif

#ifndef HAVE_NGETTEXT
#define ngettext(a,b,c) not_here("ngettext")
#define dngettext(a,b,c,d) not_here("dngettext")
#define dncgettext(a,b,c,d,e) not_here("dcngettext")
#undef ANY_MISSING
#define ANY_MISSING 1
#endif

#ifndef HAVE_BIND_TEXTDOMAIN_CODESET
#define bind_textdomain_codeset(a,b) not_here("bind_textdomain_codeset")
#undef ANY_MISSING
#define ANY_MISSING 1
#endif

#if ANY_MISSING
static int
not_here(char *s)
{
	croak("Locale::gettext::%s not implemented on this architecture", s);
	return -1;
}
#endif

#include <langinfo.h>
#include <iconv.h>

/* originally was in gi/perl-install/c/stuff.xs.pl in http://cvs.mandriva.com/cgi-bin/cvsweb.cgi/ */
SV * iconv_(char* s, char* from_charset, char* to_charset) {
  iconv_t cd = iconv_open(to_charset, from_charset);
  char* retval = s;
  SV* retval_sv;
  if (cd != (iconv_t) (-1)) {
      size_t s_len = strlen(retval);
      /* the maximum expansion when converting happens when converting
	 tscii to utf-8; each tscii char can become up to 4 unicode chars
	 and each one of those unicode chars can be 3 bytes long */
      char *buf = alloca(4 * 3 * s_len);
      {
	  char *ptr = buf;
	  size_t ptr_len = 4 * 3 * s_len;
	  if ((iconv(cd, &s, &s_len, &ptr, &ptr_len)) != (size_t) (-1)) {
	      *ptr = 0;
	      retval = buf;
	  }
      }
      iconv_close(cd);
  }
  retval_sv = newSVpv(retval, 0);
  if (!strcmp(to_charset, "UTF-8") || !strcmp(to_charset, "UTF8"))
      SvUTF8_on(retval_sv);
  return retval_sv;
}

MODULE = Locale::gettext	PACKAGE = Locale::gettext

double
constant(name,arg)
	char *		name
	int		arg

SV *
gettext(msgid)
	char *		msgid
    CODE:
	RETVAL = newSVpv( gettext(msgid), 0 );
	if (is_domain_utf8(current_domain))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

SV *
dcgettext(domainname, msgid, category)
	char *		domainname
	char *		msgid
	int		category
    CODE:
	RETVAL = newSVpv( dcgettext(domainname, msgid, category), 0 );
	if (is_domain_utf8(domainname))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

SV *
dgettext(domainname, msgid)
	char *		domainname
	char *		msgid
    CODE:
	RETVAL = newSVpv( dgettext(domainname, msgid), 0 );
	if (is_domain_utf8(domainname))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

SV *
ngettext(msgid, msgid_plural, n)
	char *		msgid
	char *		msgid_plural
	unsigned long	n
    CODE:
	RETVAL = newSVpv( ngettext(msgid, msgid_plural, n), 0 );
	if (is_domain_utf8(current_domain))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

SV *
dcngettext(domainname, msgid, msgid_plural, n, category)
	char *		domainname
	char *		msgid
	char *		msgid_plural
	unsigned long	n
	int		category
    CODE:
	RETVAL = newSVpv( dcngettext(domainname, msgid, msgid_plural, n, category), 0 );
	if (is_domain_utf8(domainname))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

SV *
dngettext(domainname, msgid, msgid_plural, n)
	char *		domainname
	char *		msgid
	char *		msgid_plural
	unsigned long	n
    CODE:
	RETVAL = newSVpv( dngettext(domainname, msgid, msgid_plural, n), 0 );
	if (is_domain_utf8(domainname))
	    SvUTF8_on(RETVAL);
    OUTPUT:
	RETVAL

char *
textdomain(domain)
	char *		domain
    CODE:
	RETVAL = textdomain(domain);
	if (current_domain) free(current_domain);
	current_domain = strdup(domain);
    OUTPUT:
	RETVAL

char *
bindtextdomain(domain, dirname = NULL)
	char *		domain
	char *		dirname

char *
bind_textdomain_codeset(domain, codeset = NULL)
	char *		domain
	char *		codeset
    CODE:
	RETVAL = bind_textdomain_codeset(domain, codeset);
	if (strcmp(codeset, "UTF8") == 0 || strcmp(codeset, "UTF-8") == 0) {	  
	    if (!domains_in_utf8) domains_in_utf8 = newHV();
	    hv_store(domains_in_utf8, domain, strlen(domain), newSV(0), 0);
	}

    OUTPUT:
	RETVAL
	 
SV *
iconv(s, from_charset, to_charset)
  char *s
  SV *from_charset
  SV *to_charset
  CODE:
  RETVAL = iconv_( s 
		 , from_charset == &PL_sv_undef ? nl_langinfo(CODESET) : SvPV_nolen(from_charset)
		 , to_charset == &PL_sv_undef ? nl_langinfo(CODESET) : SvPV_nolen(to_charset)
		 );
  OUTPUT:
  RETVAL
